
#include "asm_passes.h"
#include "../arm.h"

// Optimization passes
#include "stack_ra.h"
#include "register_allocation.h"
#include "simplify_asm.h"
#include "remove_identical_moves.h"
#include "block_placement.h"
#include "remove_redundant_ldrs.h"
#include "condify.h"
#include "use_analysis.h"
#include "algebraic_identity.h"
#include "remove_unused_instruction.h"
#include "peephole.h"
#include "remove_redundant_moves.h"

inline void run_on_every_function(Program_Asm *program_asm, Asm_Func_Pass pass) {
    for(auto func_asm : program_asm->functions) {
        pass(func_asm);
    }
}

#define RUN(pass_name) pass_name(program_asm)
void bless(Program_Asm *program_asm, bool opt) {
	RUN(remove_redundant_ldrs);
    RUN(use_analysis);
    // RUN(remove_redundant_moves);
    // RUN(use_analysis);
    RUN(algebraic_identity);
    RUN(use_analysis);
    RUN(peephole);
    RUN(use_analysis);
    RUN(remove_unused_instruction);
    RUN(use_analysis);
    register_allocation(program_asm, opt);
    RUN(remove_identical_moves);
    printf("ok\n");
}
#undef RUN

// helper functions

void replace_uses(MI *I, MOperand *old_opr, MOperand *new_opr, Func_Asm *func) {
    Array<MOperand *> oprs = get_uses_ptr(I, func->has_return_value);
    for(MOperand *opr_ptr : oprs) {
        if (opr_ptr && *opr_ptr == *old_opr) {
            // modify use
            auto use = opr_ptr->get_use_head(func);
            while(use != nullptr) {
                if(use->I == I) {
                    use->rm_use(func);
                    auto new_use = new MI_Use(new_opr, use->user, I);
                    // user is nullptr when MI_Store
                    if(use->user != nullptr)
                        use->user->add_use(new_use, func);
                }
                use = use->next;
            }
            *opr_ptr = *new_opr;
        }
    }
}

void s_replace_uses(MI *I, MOperand *old_opr, MOperand *new_opr, Func_Asm *func) {
    Array<MOperand *> oprs = s_get_uses_ptr(I, func->has_return_value);
    for(MOperand *opr_ptr : oprs) {
        if (opr_ptr && *opr_ptr == *old_opr) {
            // modify use
            auto use = opr_ptr->s_get_use_head(func);
            while(use != nullptr) {
                if(use->I == I) {
                    use->rm_use(func);
                    auto new_use = new MI_Use(new_opr, use->user, I);
                    // user is nullptr when MI_VStore
                    if(use->user != nullptr)
                        use->user->s_add_use(new_use, func);
                }
                use = use->next;
            }
            *opr_ptr = *new_opr;
        }
    }
}

void replace_defs(MI *I, MOperand *old_opr, MOperand *new_opr, Func_Asm *func) {
    Array<MOperand *> oprs = get_defs_ptr(I);
    for(MOperand *opr_ptr : oprs) {
        if (opr_ptr && *opr_ptr == *old_opr) {
            // modify use
            auto uses = get_uses_ptr(I, func->has_return_value);
            for(auto use_operand: uses) {
                rm_instruction_use(I, use_operand, func);
                auto new_use = new MI_Use(use_operand, new_opr, I);
                use_operand->add_use(new_use, func);
            }
            *opr_ptr = *new_opr;
        }
    }
}

void s_replace_defs(MI *I, MOperand *old_opr, MOperand *new_opr, Func_Asm *func) {
    Array<MOperand *> oprs = s_get_defs_ptr(I);
    for(MOperand *opr_ptr : oprs) {
        if (opr_ptr && *opr_ptr == *old_opr) {
            // modify use
            auto uses = s_get_uses_ptr(I, func->has_return_value);
            for(auto use_operand: uses) {
                s_rm_instruction_use(I, use_operand, func);
                auto new_use = new MI_Use(use_operand, new_opr, I);
                use_operand->s_add_use(new_use, func);
            }
            *opr_ptr = *new_opr;
        }
    }
}

Array<MOperand> get_defs(MI *I) {
    Array<MOperand> oprs;
    switch(I->tag) {
        case MI_MOVE:   { oprs.push(((MI_Move *)I)->dst); } break;
        case MI_BINARY: { oprs.push(((MI_Binary *)I)->dst); } break;
        case MI_FUNC_CALL: {
            oprs.push(make_reg(ra));
            for(uint8 r = x1; r < REG_COUNT; r++) {
                if (is_caller_save(r)) {
                    if (r == ra) cout << "ra appear in function call" << endl;
                    oprs.push(make_reg(r));
                }
            }
        } break;

        case MI_LOAD: { oprs.push(((MI_Load *)I)->reg); } break;

        case MI_CLZ: { oprs.push(((MI_Clz *)I)->dst); } break;

        case MI_SLT: { oprs.push(((MI_Slt *)I)->dst); } break;

        case MI_VCVT: {
            auto fcvt = (MI_VCvt *)I;
            if(fcvt->to_type == S32) {
                oprs.push(fcvt->dst);
            }
        } break;

        case MI_VCMP_SET: {
            oprs.push(((MI_VCmp_Set *)I)->dst);
        } break;

        case MI_COMPARE:
        case MI_PUSH:
        case MI_RETURN:
        case MI_POP:
        case MI_STORE:
        case MI_BRANCH: 
        //----float----
        case MI_VMOVE:
        case MI_VBINARY:
        case MI_VPUSH:
        case MI_VPOP:
        case MI_VLOAD: // vldr
        case MI_VSTORE:  // vstr
        break;
        
        default: {
            printf("unknown instrID: %d\n", I->tag);
            assert(false && "meet unknown instr in get_defs");
        }
    }
    return oprs;
}

Array<MOperand> get_uses(MI *I, bool func_has_return_value) {
    Array<MOperand> oprs;

    switch(I->tag) {

        case MI_MOVE:   {
            if(((MI_Move *)I)->src.tag == REG || ((MI_Move *)I)->src.tag == VREG){
                oprs.push(((MI_Move *)I)->src);
            }
        } break;

        case MI_VMOVE: {
            if(((MI_VMove *)I)->src.tag == REG || ((MI_VMove *)I)->src.tag == VREG){
                oprs.push(((MI_VMove *)I)->src);
            }
        } break;
        case MI_CLZ:   {
            oprs.push(((MI_Clz *)I)->operand);
        } break;

        case MI_BINARY: {
            oprs.push(((MI_Binary *)I)->lhs);
            if(((MI_Binary *)I)->rhs.tag != IMM)
                oprs.push(((MI_Binary *)I)->rhs);
        } break;

        case MI_COMPARE: {
            oprs.push(((MI_Compare *)I)->lhs);
            oprs.push(((MI_Compare *)I)->rhs);
        } break;

        case MI_SLT: {
            oprs.push(((MI_Slt *)I)->lhs);
            oprs.push(((MI_Slt *)I)->rhs);
        } break;

        case MI_FUNC_CALL: {
            auto call = (MI_Func_Call *)I;
            // @TODO: get uses based on arg count
            // @TODO: temporarily use 4 argument regs to keep compatibility
            for(uint8 r = a0; r <= a3; r++) {
                if (r >= call->arg_count + a0) break;
                oprs.push(make_reg(r));
            }
        } break;

        case MI_STORE: { 
            auto store = (MI_Store *)I;
            oprs.push(store->reg);
            if(store->base.tag == VREG || store->base.tag == REG)
                oprs.push(store->base);
            if(store->offset.tag == VREG || store->offset.tag == REG)
                oprs.push(store->offset);
        } break;

        case MI_LOAD: { 
            auto load = (MI_Load *)I;
            if(load->base.tag == VREG || load->base.tag == REG)
                oprs.push(load->base);
            if(load->offset.tag == VREG || load->offset.tag == REG)
                oprs.push(load->offset);
        } break;

        case MI_PUSH: {
            auto push = (MI_Push *) I;
            for (auto push_operand : push->operands) {
                oprs.push(push_operand);
            }
        } break;

        case MI_RETURN: {
            oprs.push(make_reg(ra));
            if (func_has_return_value) {
                oprs.push(make_reg(a0));
            }
        } break;

        case MI_VLOAD:{
            auto vldr = (MI_VLoad *)I;
            if(vldr->base.tag != ERRORTYPE)
                oprs.push(vldr->base);
            if(vldr->offset.tag != ERRORTYPE)
                oprs.push(vldr->offset);
        } break;
        case MI_VSTORE:{
            auto vstr = (MI_VStore *)I;
            if(vstr->base.tag != ERRORTYPE)
                oprs.push(vstr->base);
            if(vstr->offset.tag != ERRORTYPE)
                oprs.push(vstr->offset);
        } break;
        case MI_BRANCH: {
            auto br_I = (MI_Branch *)I;
            if (br_I->cond != NO_CONDITION) {
                oprs.push(br_I->lhs);
                oprs.push(br_I->rhs);
            }
        } break;
        case MI_VCVT: {
            auto fcvt = (MI_VCvt *)I;
            if(fcvt->from_type == S32) {
                oprs.push(fcvt->src);
            }
        } break;

        case MI_POP:
        
        //----float----(return nothing in get_uses())
        case MI_VCMP_SET:
        case MI_VBINARY:
        case MI_VPUSH:
        case MI_VPOP:
        break;
        default: {
            printf("unknown instrID: %d\n", I->tag);
            assert(false && "meet unknown instr in get_uses");
        }
    }

    return oprs;
}

Array<MOperand> s_get_defs(MI *I) {
    Array<MOperand> oprs;
    switch(I->tag) {
        case MI_VMOVE:   { oprs.push(((MI_VMove *)I)->dst); } break;
        case MI_VBINARY: { oprs.push(((MI_VBinary *)I)->dst); } break;

        case MI_VLOAD:   { oprs.push(((MI_VLoad *)I)->reg); } break;

        case MI_VCMP_SET: { oprs.push(((MI_VCmp_Set *)I)->dst); } break;

        case MI_VCVT: {
            auto fcvt = (MI_VCvt *)I;
            if(fcvt->to_type == F32) {
                oprs.push(fcvt->dst);
            }
        } break;

        case MI_FUNC_CALL: {
            for(uint8 r = f0; r < SREG_COUNT; r++) {
                if (s_is_caller_save(r)) {
                    oprs.push(make_sreg(r));
                }
            }
        } break;

        case MI_VPUSH:
        case MI_VPOP:
        case MI_VSTORE:
        case MI_MOVE:
        case MI_BINARY:
        case MI_CLZ:
        case MI_RETURN:
        case MI_BRANCH:
        case MI_COMPARE:
        case MI_PUSH:
        case MI_POP:
        case MI_LOAD:
        case MI_STORE:
        case MI_SLT:
        break;
        default: {
            printf("unknown instrID: %d\n", I->tag);
            assert(false && "meet unknown instr in s_get_defs");
        }
    }
    return oprs;
}

Array<MOperand> s_get_uses(MI *I, bool func_has_return_value) {
    Array<MOperand> oprs;

    switch(I->tag) {

        case MI_VMOVE:   {
            if (!(((MI_VMove *)I)->from_s32)) {
                oprs.push(((MI_VMove *)I)->src);
            }
        } break;

        case MI_VBINARY: {
            oprs.push(((MI_VBinary *)I)->lhs);
            if (((MI_VBinary *)I)->fneg == 0) {
                oprs.push(((MI_VBinary *)I)->rhs);
            }
        } break;

        case MI_VCMP_SET: {
            oprs.push(((MI_VCmp_Set *)I)->lhs);
            oprs.push(((MI_VCmp_Set *)I)->rhs);
        } break;

        // 'base' and 'offset' are vreg not vsreg
        case MI_VSTORE: { 
            oprs.push(((MI_VStore *)I)->reg);
        } break;

        case MI_VLOAD: {} break;

        case MI_VCVT: {
            auto fcvt = (MI_VCvt *)I;
            if(fcvt->from_type == F32) {
                oprs.push(fcvt->src);
            }
        } break;

        case MI_FUNC_CALL: {
            for(uint8 r = f10; r < f14; r++) {
                oprs.push(make_sreg(r));
            }
        } break;

        case MI_RETURN: {
            if (func_has_return_value) {
                oprs.push(make_sreg(f10));
            }
        } break;

        case MI_SLT:
        case MI_VPOP:
        case MI_VPUSH:
        case MI_MOVE:
        case MI_BINARY:
        case MI_CLZ:
        case MI_BRANCH:
        case MI_COMPARE:
        case MI_PUSH:
        case MI_POP:
        case MI_LOAD:
        case MI_STORE:
        break;
        default: {
            printf("unknown instrID: %d\n", I->tag);
            assert(false && "meet unknown instr in s_get_uses");
        }

    }

    return oprs;
}

Array<MOperand *> get_defs_ptr(MI *I) {
    Array<MOperand *> oprs;
    switch(I->tag) {
        case MI_MOVE:   { oprs.push(&(((MI_Move *)I)->dst)); } break;
        case MI_BINARY: { oprs.push(&(((MI_Binary *)I)->dst)); } break;
        case MI_FUNC_CALL: {} break;

        case MI_LOAD: { oprs.push(&(((MI_Load *)I)->reg)); } break;

        case MI_CLZ: { oprs.push(&(((MI_Clz *)I)->dst)); } break;

        case MI_SLT: { oprs.push(&((MI_Slt *)I)->dst); } break;

        case MI_VCVT: {
            auto fcvt = (MI_VCvt *)I;
            if(fcvt->to_type == S32) {
                oprs.push(&fcvt->dst);
            }
            else {
                oprs.push(nullptr);
            }
        } break;

        case MI_VCMP_SET: {
            oprs.push(&((MI_VCmp_Set *)I)->dst);
        } break;

        // use 'nullptr' as a placeholder
        case MI_COMPARE:
        case MI_PUSH:
        case MI_RETURN:
        case MI_POP:
        case MI_BRANCH: 
        case MI_STORE: {
            oprs.push(nullptr);
        } break;

        //----float----
        case MI_VMOVE:
        case MI_VBINARY:
        case MI_VPUSH:
        case MI_VPOP:
        case MI_VLOAD:
        case MI_VSTORE:
        break;
        
        default: {
            printf("unknown instrID: %d\n", I->tag);
            assert(false && "meet unknown instr in get_defs");
        }
    }
    return oprs;
}

Array<MOperand *> get_uses_ptr(MI *I, bool func_has_return_value) {
    Array<MOperand *> oprs;

    switch(I->tag) {

        case MI_MOVE:   {
            if(((MI_Move *)I)->src.tag == REG || ((MI_Move *)I)->src.tag == VREG){
                oprs.push(&((MI_Move *)I)->src);
            }
        } break;

        case MI_VMOVE: {
            if(((MI_VMove *)I)->src.tag == REG || ((MI_VMove *)I)->src.tag == VREG){
                oprs.push(&((MI_VMove *)I)->src);
            }
        } break;
        case MI_CLZ:   {
            oprs.push(&((MI_Clz *)I)->operand);
        } break;

        case MI_BINARY: {
            oprs.push(&((MI_Binary *)I)->lhs);
            oprs.push(&((MI_Binary *)I)->rhs);
        } break;

        case MI_COMPARE: {
            oprs.push(&((MI_Compare *)I)->lhs);
            oprs.push(&((MI_Compare *)I)->rhs);
        } break;

        case MI_SLT: {
            oprs.push(&((MI_Slt *)I)->lhs);
            oprs.push(&((MI_Slt *)I)->rhs);
        } break;

        case MI_FUNC_CALL: {} break;

        case MI_STORE: { 
            auto store = (MI_Store *)I;
            oprs.push(&store->reg);
            if(store->base.tag != ERRORTYPE)
                oprs.push(&store->base);
            if(store->offset.tag != ERRORTYPE)
                oprs.push(&store->offset);
        } break;

        case MI_LOAD: { 
            auto load = (MI_Load *)I;
            if(load->base.tag != ERRORTYPE)
                oprs.push(&load->base);
            if(load->offset.tag != ERRORTYPE)
                oprs.push(&load->offset);
        } break;

        case MI_PUSH: {} break;

        case MI_RETURN: {} break;

        case MI_VLOAD:{
            auto vldr = (MI_VLoad *)I;
            if(vldr->base.tag != ERRORTYPE)
                oprs.push(&vldr->base);
            if(vldr->offset.tag != ERRORTYPE)
                oprs.push(&vldr->offset);
        } break;
        case MI_VSTORE:{
            auto vstr = (MI_VStore *)I;
            if(vstr->base.tag != ERRORTYPE)
                oprs.push(&vstr->base);
            if(vstr->offset.tag != ERRORTYPE)
                oprs.push(&vstr->offset);
        } break;
        case MI_BRANCH: {
            auto br_I = (MI_Branch *)I;
            if (br_I->cond != NO_CONDITION) {
                oprs.push(&br_I->lhs);
                oprs.push(&br_I->rhs);
            }
        } break;
        case MI_VCVT: {
            auto fcvt = (MI_VCvt *)I;
            if(fcvt->from_type == S32) {
                oprs.push(&fcvt->src);
            }
        } break;

        case MI_POP:
        
        //----float----(return nothing in get_uses())
        case MI_VCMP_SET:
        case MI_VBINARY:
        case MI_VPUSH:
        case MI_VPOP:
        break;
        default: {
            printf("unknown instrID: %d\n", I->tag);
            assert(false && "meet unknown instr in get_uses");
        }
    }

    return oprs;
}

Array<MOperand *> s_get_defs_ptr(MI *I) {
    Array<MOperand *> oprs;
    switch(I->tag) {
        case MI_VMOVE:   { oprs.push(&((MI_VMove *)I)->dst); } break;
        case MI_VBINARY: { oprs.push(&((MI_VBinary *)I)->dst); } break;

        case MI_VLOAD:   { oprs.push(&((MI_VLoad *)I)->reg); } break;

        case MI_VCMP_SET: { oprs.push(&((MI_VCmp_Set *)I)->dst); } break;

        case MI_VCVT: {
            auto fcvt = (MI_VCvt *)I;
            if(fcvt->to_type == F32) {
                oprs.push(&fcvt->dst);
            }
            else {
                oprs.push(nullptr);
            }
        } break;

        case MI_FUNC_CALL: {} break;

        case MI_VPUSH:
        case MI_VPOP:
        case MI_VSTORE: {
            // use nullptr as a placeholder
            oprs.push(nullptr);
        } break;

        case MI_MOVE:
        case MI_BINARY:
        case MI_CLZ:
        case MI_RETURN:
        case MI_BRANCH:
        case MI_COMPARE:
        case MI_PUSH:
        case MI_POP:
        case MI_LOAD:
        case MI_STORE:
        case MI_SLT:
        break;
        default: {
            printf("unknown instrID: %d\n", I->tag);
            assert(false && "meet unknown instr in s_get_defs");
        }
    }
    return oprs;
}

Array<MOperand *> s_get_uses_ptr(MI *I, bool func_has_return_value) {
    Array<MOperand *> oprs;

    switch(I->tag) {

        case MI_VMOVE:   {
            if (!(((MI_VMove *)I)->from_s32)) {
                oprs.push(&((MI_VMove *)I)->src);
            }
        } break;

        case MI_VBINARY: {
            oprs.push(&((MI_VBinary *)I)->lhs);
            if (((MI_VBinary *)I)->fneg == 0) {
                oprs.push(&((MI_VBinary *)I)->rhs);
            }
        } break;

        case MI_VCMP_SET: {
            oprs.push(&((MI_VCmp_Set *)I)->lhs);
            oprs.push(&((MI_VCmp_Set *)I)->rhs);
        } break;

        // 'base' and 'offset' are vreg not vsreg
        case MI_VSTORE: { 
            oprs.push(&((MI_VStore *)I)->reg);
        } break;

        case MI_VLOAD: {} break;

        case MI_VCVT: {
            auto fcvt = (MI_VCvt *)I;
            if(fcvt->from_type == F32) {
                oprs.push(&fcvt->src);
            }
        } break;

        case MI_FUNC_CALL: {} break;

        case MI_RETURN: {} break;

        case MI_SLT:
        case MI_VPOP:
        case MI_VPUSH:
        case MI_MOVE:
        case MI_BINARY:
        case MI_CLZ:
        case MI_BRANCH:
        case MI_COMPARE:
        case MI_PUSH:
        case MI_POP:
        case MI_LOAD: // ldr
        case MI_STORE: // str
        break;
        default: {
            printf("unknown instrID: %d\n", I->tag);
            assert(false && "meet unknown instr in s_get_uses");
        }

    }

    return oprs;
}

// delete uses that 'opr' uses
void delete_user(MOperand *opr, Func_Asm *func) {
    auto I = opr->get_def_I(func);
    if(I != nullptr) {
        auto use_operands = get_uses(I, func->has_return_value);
        for(auto use_operand: use_operands) {
            auto use = use_operand.get_use_head(func);
            while(use != nullptr) {
                if(use->I == I) {
                    use->rm_use(func);
                }
                use = use->next;
            }
        }
    }
}

// replace all uses of 'opr' by 'new_opr'
void replace_operand_by_operand(MOperand *opr, MOperand *new_opr, Func_Asm *func) {
    if(opr == nullptr || new_opr == nullptr)
        return;
    auto use = opr->get_use_head(func);
    while(use != nullptr) {
        auto user = use->user;
        assert(use->I != nullptr);
        replace_uses(use->I, opr, new_opr, func);
        use = use->next;
    }
    delete_user(opr, func);
}

