#include "Block.h"
#include "Loop.h"


Block::Block() : regNum{0}
{
    basicBlocks.emplace_back(BasicBlockPtr(new BasicBlock()));
}

void Block::setBBbelongFunc(shared_ptr<Function> func){
    for(auto bb:basicBlocks){
        bb->belongfunc = func;
    }
}

void Block::getLabelBBMap(){
    for(int i=0;i<basicBlocks.size();i++){
        LabelBBMap[basicBlocks[i]->label] = basicBlocks[i];
    }
}

void Block::pushVariable(VariablePtr variable)
{
    variables[variable->name]= variable;
}

VariablePtr Block::findVariable(string name)
{
    if (variables.find(name)!=variables.end()){
        return variables[name];
    }

    return nullptr;
}

void Block::allocLocalVariable()
{
    auto tmp = basicBlocks[0]->instructions;
    basicBlocks[0]->instructions.clear();
    for (auto &var : variables) basicBlocks[0]->instructions.emplace_back(InstructionPtr(new AllocaInstruction(var.second, basicBlocks[0])));
    for(auto &ins: tmp) basicBlocks[0]->instructions.emplace_back(ins);
}

void Block::solveEndInstruction()
{
    for (auto &basicBlock : basicBlocks)
    {
        vector<shared_ptr<Instruction>> newIns;
        int i  =0;
        for(;i<basicBlock->instructions.size();i++){
            // cerr<<i<<"  "<<basicBlock->instructions[i]->type<<endl;
            newIns.push_back(basicBlock->instructions[i]);
            if(basicBlock->instructions[i]->type==Store&&dynamic_cast<StoreInstruction*>(basicBlock->instructions[i].get())->des->name=="retval"){
                break;
            }
            
        }
        basicBlock->instructions = newIns;

        assert(basicBlock->endInstruction);
        basicBlock->instructions.emplace_back(basicBlock->endInstruction);
    }
}

void Block::pushBasicBlock(BasicBlockPtr basicblock)
{
    basicBlocks.emplace_back(basicblock);
}

void Block::print()
{
    for (int i = 0; i < basicBlocks.size(); i++)
    {
        basicBlocks[i]->print();
        if (i != basicBlocks.size() - 1)
            cout << endl;
    }
}

void BasicBlock::pushInstruction(InstructionPtr instruction)
{
    if(!endInstruction) instructions.emplace_back(instruction);
    instruction->basicblock = this->instructions[0]->basicblock;
}

void BasicBlock::pushInstruction(vector<InstructionPtr> toInsert)
{
    auto tmpEnd = instructions.empty() ? nullptr : instructions.back();
    vector<InstructionPtr> newInstr(instructions);
    auto bb = instructions[0]->basicblock;

    if(tmpEnd) {
        assert(tmpEnd->type == Br || tmpEnd->type == Return);
        newInstr.pop_back();
        endInstruction = nullptr;
    }
    for(auto instr: toInsert) {
        instr->basicblock = bb;
        newInstr.emplace_back(instr);
    }
    if(tmpEnd) {
        newInstr.emplace_back(tmpEnd);
        endInstruction = tmpEnd;
    }
    instructions = newInstr;
}

void BasicBlock::pushFrontInstruction(vector<InstructionPtr> toInsert)
{
    assert(instructions.size() > 0);
    auto bb = instructions[0]->basicblock;
    for(auto instr: toInsert) {
        instr->basicblock = bb;
    }
    for(auto instr: instructions) {
        toInsert.emplace_back(instr);
    }
    // toInsert.emplace_back(instructions.begin(), instructions.end());
    instructions = toInsert;
}

// insert an instruction before the second
void BasicBlock::insertInstruction(InstructionPtr instruction, InstructionPtr next)
{
    auto curInstruction = instructions.begin();
    while (curInstruction != instructions.end())
    {
        if (*curInstruction == next) break;
        curInstruction++;
    }
    // make sure next can be found
    assert(curInstruction != instructions.end());
    instructions.emplace(curInstruction, instruction);
    // instruction->basicblock = this->instructions[0]->basicblock;
    instruction->basicblock = next->basicblock;
}

void BasicBlock::removeInsturction(InstructionPtr instruction)
{
    assert(instruction != endInstruction);
    auto curInstruction = instructions.begin();
    int time = 0;
    while (curInstruction != instructions.end())
    {
        // cerr << time++ << endl;
        if (*curInstruction == instruction) break;
        curInstruction++;
    }
    // cerr << "[passed] loop" << endl;
    // make sure next can be found
    assert(curInstruction != instructions.end());
    instructions.erase(curInstruction);
}

void BasicBlock::pushEndInstruction(InstructionPtr instruction)
{
    if(!endInstruction) endInstruction = instruction;
}

void BasicBlock::print()
{
    label->print();

    int cnt = 0,n=predBasicBlocks.size();
    for(auto &pred : predBasicBlocks){
        if(cnt==0){
            cout<<"\t\t\t; preds = ";
        }
        cout<<"%"+pred->label->name;
        if(cnt<n-1){
            cout<<", ";
        }
        cnt++;
    }
    cout<<endl;
    for (auto &instruction : instructions)
    {
        instruction->print();
    }
}

void Block::clearBBToLoop() {
    bbToLoop.clear();
    for(auto basicBlock: basicBlocks) {
        basicBlock->loop = nullptr;
        basicBlock->loopDepth = 0;
    }
}

void Block::clearLoops() {
    loops.clear();
    topLoops.clear();
}

BasicBlockPtr Block::getEntryBlock() {
    assert(basicBlocks.size() > 0);
    return basicBlocks[0];
}

void BasicBlock::basicBlockDFSPost(shared_ptr<BasicBlock> bb, function<bool(shared_ptr<BasicBlock>)> cond, function<void(shared_ptr<BasicBlock>)> action) {
    if(cond(bb))
        return;
    for(auto succ : bb->getSuccessor())
        basicBlockDFSPost(succ, cond, action);
    action(bb);
}

void BasicBlock::domTreeDFSPost(shared_ptr<BasicBlock> bb, function<bool(shared_ptr<BasicBlock>)> cond, function<void(shared_ptr<BasicBlock>)> action) {
    if(cond(bb))
        return;
    for(auto succ : bb->getDominatorSon())
        domTreeDFSPost(succ, cond, action);
    action(bb);
}

void preorderVisit(LoopPtr loop, set<LoopPtr>& visited, vector<LoopPtr>& preorder, int& loopCnt) {
    visited.insert(loop);
    for(auto innerLoop: loop->getSubLoops()) {
        preorderVisit(innerLoop, visited, preorder, loopCnt);
    }
    preorder[loopCnt ++] = loop;
}

// preorder traverse loops and their subloops
vector<LoopPtr> Block::getLoopsInPostorder() {
    set<LoopPtr> visited;
    vector<LoopPtr> preorder(loops.size());
    int loopCnt = 0;
    for(auto loop: loops) {
        if(visited.find(loop) != visited.end()) continue;
        preorderVisit(loop, visited, preorder, loopCnt);
    }
    assert(loopCnt == loops.size());
    return preorder;
}

void BasicBlock::rankReorder()
{
    rank.clear();
    int r = 0;
    for(auto instr : instructions)
    {
        rank[instr.get()] = r++;
    }
}

BasicBlockPtr lcaCompute(shared_ptr<BasicBlock> bb1, shared_ptr<BasicBlock> bb2)
{
    int rank1 = 0, rank2 = 0;
    BasicBlockPtr now1 = bb1, now2 = bb2;
    while(now1)
    {
        rank1++;
        now1 = now1->directDominator;
    } 
    while(now2)
    {
        rank2++;
        now2 = now2->directDominator;
    }
    now1 = bb1, now2 = bb2;
    if(rank1 >= rank2)
    {
        while(rank1 > rank2)
        {
            now1 = now1->directDominator;
            rank1--;
        }
    }
    else{
        while(rank2 > rank1)
        {
            now2 = now2->directDominator;
            rank2--;
        }
    }
    return nullptr;
} 