#pragma once
#include <unordered_map>
#include <unordered_set>
#include <iostream>
#include <set>
#include <stack>
#include <memory>
#include <functional>
#include "Variable.h"
#include "Instruction.h"
#include "Label.h"
#include <algorithm>
#include <map>

using namespace std;
struct BasicBlock;
struct Instruction;
struct PhiInstruction;
struct IcmpInstruction;
struct Function;
struct Loop;

struct Block
{
    int regNum;
    ValuePtr getReg(TypePtr type){return ValuePtr(new Reg(type, regNum++));}
    vector<shared_ptr<BasicBlock>> basicBlocks;
    std::unordered_map<string, VariablePtr> variables;

    unordered_map<LabelPtr, shared_ptr<BasicBlock>> LabelBBMap;
    void getLabelBBMap();

    // loops
    vector<shared_ptr<Loop>> loops;
    unordered_map<shared_ptr<BasicBlock>, shared_ptr<Loop>> bbToLoop;
    vector<shared_ptr<Loop>> topLoops;
    vector<shared_ptr<Loop>> getAllLoops() { return loops; }
    vector<shared_ptr<Loop>> getLoopsInPostorder();
    vector<shared_ptr<Loop>> getTopLoops() { return topLoops; }
    void addLoop(shared_ptr<Loop> loop) { loops.push_back(loop); }
    void addTopLoop(shared_ptr<Loop> loop) { topLoops.push_back(loop); }
    void addBBToLoop(shared_ptr<BasicBlock> bb, shared_ptr<Loop> loop)  { bbToLoop[bb] = loop; }
    shared_ptr<Loop> getLoopOfBB(shared_ptr<BasicBlock> bb) { return bbToLoop.count(bb)? bbToLoop[bb]: nullptr; }
    void clearBBToLoop();
    void clearLoops();

    Block();

    void setBBbelongFunc(shared_ptr<Function> func);
    void pushBasicBlock(shared_ptr<BasicBlock> block);
    void pushVariable(VariablePtr variable);
    VariablePtr findVariable(string name);
    void allocLocalVariable();
    void solveEndInstruction();
    shared_ptr<BasicBlock> getEntryBlock();
    void print();
};
typedef shared_ptr<Block> BlockPtr;

struct BasicBlock
{
    LabelPtr label;
    vector<shared_ptr<Instruction>> instructions;
    map<Instruction*, int> rank;
    shared_ptr<Instruction> endInstruction= nullptr;
    shared_ptr<Function> belongfunc;

    BasicBlock(LabelPtr label=LabelPtr(new Label())): label{label} ,belongfunc{nullptr}{};
    void setBelongFunc(shared_ptr<Function> func){
        belongfunc = func;
    }
    shared_ptr<Function> getParent() { return belongfunc; }


    void pushInstruction(shared_ptr<Instruction> instruction);
    void pushInstruction(vector<shared_ptr<Instruction>> toInsert);
    void insertInstruction(shared_ptr<Instruction> instruction, shared_ptr<Instruction> next);
    void pushEndInstruction(shared_ptr<Instruction> instruction);
    void pushFrontInstruction(vector<shared_ptr<Instruction>> toInsert);
    void removeInsturction(shared_ptr<Instruction> instruction);

    //用于后续优化阶段的标记
    bool isDeadBB=false;
    bool isWhileCond = false;
    bool isIfCond = false;

    int loopDepth = 0;
    shared_ptr<Loop> loop = nullptr;
    static void basicBlockDFSPost(shared_ptr<BasicBlock> bb, function<bool(shared_ptr<BasicBlock>)> cond, function<void(shared_ptr<BasicBlock>)> action);
    static void domTreeDFSPost(shared_ptr<BasicBlock> bb, function<bool(shared_ptr<BasicBlock>)> cond, function<void(shared_ptr<BasicBlock>)> action);

    //前继后继基本块
    unordered_set<shared_ptr<BasicBlock>> predBasicBlocks;
    unordered_set<shared_ptr<BasicBlock>> succBasicBlocks;
    
    //该基本块直接支配的子基本块
    unordered_set<shared_ptr<BasicBlock>> dominatorSon;
    //支配边界
    unordered_set<shared_ptr<BasicBlock>> DF;
    //用来存储直接支配该基本块的基本块
    shared_ptr<BasicBlock> directDominator = nullptr;
    //能够支配的该基本块所有基本块
    unordered_set<shared_ptr<BasicBlock>> allDoms;

    void addPredBasicBlock(shared_ptr<BasicBlock> bb)       { predBasicBlocks.insert(bb); }
    void addSuccBasicBlock(shared_ptr<BasicBlock> bb)       { succBasicBlocks.insert(bb); }
    void deletePredBasicBlock(shared_ptr<BasicBlock> bb)    { predBasicBlocks.erase(bb); }
    void deleteSuccBasicBlock(shared_ptr<BasicBlock> bb)    { succBasicBlocks.erase(bb); }
    unordered_set<shared_ptr<BasicBlock>> getPredecessor()  { return predBasicBlocks; }
    unordered_set<shared_ptr<BasicBlock>> getSuccessor()    { return succBasicBlocks; }
    int getIndegree()                                       { return predBasicBlocks.size(); }
    int getOutdegree()                                      { return succBasicBlocks.size(); }

    void addDominatorSon(shared_ptr<BasicBlock> bb)         { dominatorSon.insert(bb); }
    void addDF(shared_ptr<BasicBlock> bb)                   { DF.insert(bb); }
    void deleteDominatorSon(shared_ptr<BasicBlock> bb)      { dominatorSon.erase(bb); }
    void deleteDF(shared_ptr<BasicBlock> bb)                { DF.erase(bb); }
    unordered_set<shared_ptr<BasicBlock>> getDominatorSon() { return dominatorSon; }
    unordered_set<shared_ptr<BasicBlock>> getDF()           { return DF; }
    void setDirectDominator(shared_ptr<BasicBlock> bb)      { directDominator = bb; }
    void setAllDoms(unordered_set<shared_ptr<BasicBlock>> doms) { allDoms = doms; }
    shared_ptr<BasicBlock> getDirectDominator()             { return directDominator; }
    unordered_set<shared_ptr<BasicBlock>> getAllDoms()      { return allDoms; }
    void clearAllDoms()                                     { allDoms.clear(); }

    void print();
    void rankReorder();
};

shared_ptr<BasicBlock> lcaCompute(shared_ptr<BasicBlock> bb1, shared_ptr<BasicBlock> bb2);
typedef shared_ptr<BasicBlock> BasicBlockPtr;

