#pragma once
#include <cassert>
#include <memory>
#include "Variable.h"
#include "Function.h"
#include "Label.h"
#include "Block.h"


#include <memory>
#include <stdlib.h>
#include <map>
#include <iostream>

struct Function;
struct BasicBlock;
struct Instruction;
struct LoopGEPInstruction;
using std::enable_shared_from_this; // !!!

enum InsID
{
    Return,
    Br,
    Alloca,
    Store,
    Load,
    Call,
    Bitcast,
    Ext,
    Sitofp,
    Fptosi,
    GEP,
    Binary,
    Fneg,
    Icmp,
    Fcmp,
    Phi,
    LoopGEP,
    Select
};

enum IcmpKind {
    ICmpEQ,
    ICmpNE,
    ICmpSLT,
    ICmpSGT,
    ICmpSLE,
    ICmpSGE
};

struct Instruction: public enable_shared_from_this<Instruction> {
    InsID type;
    shared_ptr<BasicBlock> basicblock;
    ValuePtr reg;

    Instruction(InsID type, shared_ptr<BasicBlock> bb, ValuePtr reg = nullptr): type{type}, basicblock{bb}, reg{reg} {}
    virtual ~Instruction(){}
    // virtual void print() = 0;

    // virtual bool replaceValue(ValuePtr target, ValuePtr newValue) = 0;
    // virtual unsigned getNumOperands() = 0;
    // virtual ValuePtr getOperand(unsigned i) = 0;
    virtual void print() {}

    virtual bool replaceValue(ValuePtr target, ValuePtr newValue) { return false; }
    virtual unsigned getNumOperands() { return 0; }
    virtual ValuePtr getOperand(unsigned i) { return nullptr; }
    

    bool isCommutative = false;
    bool isAssociative = false;
    bool isIdempotent = false;
    bool isNilpotent = false;
    //O(n)的时间复杂度，后续编译时间不足可以优化
    void moveBefore(shared_ptr<Instruction> I, shared_ptr<Instruction> This);

    void replaceAllUsesWith(ValuePtr V);
    void replaceAllUsesWith(shared_ptr<Instruction> I);
    


    //获得在basicblock的instructions列表中的迭代器
    vector<shared_ptr<Instruction>>::iterator getIterator();


    void setName(string newName);
    string getName();
    string getNameForTest();
    void deleteSelfInBB();
    shared_ptr<Instruction> getSharedThis() { return shared_from_this(); }
};
typedef shared_ptr<Instruction> InstructionPtr;

struct ReturnInstruction:public Instruction {
    ValuePtr retValue;
    ReturnInstruction(ValuePtr retValue, shared_ptr<BasicBlock> bb) : Instruction{InsID::Return, bb}, retValue{retValue} {
        newUse(retValue.get(), this);
    }; // to-do: 改为 exp
    ~ReturnInstruction();

    virtual void print() override;
    virtual bool replaceValue(ValuePtr target, ValuePtr newValue) override;

    virtual unsigned getNumOperands() override;
    virtual ValuePtr getOperand(unsigned i) override;
};

struct AllocaInstruction:public Instruction {
    ValuePtr des;
    AllocaInstruction(ValuePtr des, shared_ptr<BasicBlock> bb) : Instruction{InsID::Alloca, bb}, des{des} {
        des->I = this;
    };
    ~AllocaInstruction();

    virtual void print() override;
    virtual bool replaceValue(ValuePtr target, ValuePtr newValue) override;

    virtual unsigned getNumOperands() override;
    virtual ValuePtr getOperand(unsigned i) override;
};

struct GetElementPtrInstruction:public Instruction {
    static int arrayIdxNum;
    static int arrayElementNum;
    static ValuePtr getArrayIdxReg(TypePtr type) { return ValuePtr(new Reg(type, "arrayidx" + to_string(arrayIdxNum++))); }
    static ValuePtr getArrayElementReg(TypePtr type) { return ValuePtr(new Reg(type, "arrayinit.element" + to_string(arrayElementNum++))); }
    ValuePtr from;
    vector<ValuePtr> index;
    bool isConstIdx;
    LoopGEPInstruction *loopGEP = nullptr;
    vector<ValuePtr> allIndex;
    ValuePtr base;
    ~GetElementPtrInstruction();
    GetElementPtrInstruction(ValuePtr from, vector<ValuePtr> index, shared_ptr<BasicBlock> bb) : Instruction{InsID::GEP, bb}, from{from}, index{index}, isConstIdx{false}, loopGEP(nullptr)
    {
        if (index.size() == 1)
        {
            reg = getArrayElementReg(from->type);
            newUse(from.get(), this, reg.get());
            newUse(index[0].get(), this, reg.get()); // bug
        }
        else
        {
            // std::cerr<<index.size()<<"\n";
            auto curr = from->type;
            for (int i = 1; i < index.size(); i++)
            {
                assert(curr->isArr());
                // std::cerr<<dynamic_cast<ArrType *>(curr.get())->getStr()<<"\n";
                curr = dynamic_cast<ArrType *>(curr.get())->inner;
                
                newUse(index[i].get(), this, reg.get());
                // }else if(from->type->isPtr()){
                //     reg = getArrayIdxReg(dynamic_cast<PtrType*>(from->type.get())->inner);
                // else
                // {
                //     std::cerr << from->getTypeStr() << " " << from->type->isPtr() << endl;
                //     reg = getArrayElementReg(from->type);
                // }
            }
            reg = getArrayIdxReg(curr);
            newUse(from.get(), this, reg.get());
            newUse(index[0].get(), this, reg.get()); // bug
        }
        reg->I = this;
        clearAllIndex();
        calulateAllIndex();
    }
    virtual void print() override;
    virtual bool replaceValue(ValuePtr target, ValuePtr newValue) override;
    virtual unsigned getNumOperands() override;
    virtual ValuePtr getOperand(unsigned i) override;

    void clearAllIndex();
    void calulateAllIndex();
};

struct StoreInstruction:public Instruction {
    ValuePtr des;
    ValuePtr value;
    shared_ptr<GetElementPtrInstruction> gep;
    ~StoreInstruction();
    StoreInstruction(shared_ptr<GetElementPtrInstruction> gep, ValuePtr value, shared_ptr<BasicBlock> bb) : Instruction{InsID::Store, bb}, gep{gep}, des{gep->reg}, value{value} {
        newUse(gep->reg.get(), this);
        // newUse(gep->from.get(), this);
        newUse(value.get(), this);
    };
    
    StoreInstruction(ValuePtr des, ValuePtr value, shared_ptr<BasicBlock> bb) : Instruction{InsID::Store, bb}, des{des}, value{value} {
        newUse(des.get(), this);
        newUse(value.get(), this);
    };
    virtual void print() override;
    virtual bool replaceValue(ValuePtr target, ValuePtr newValue) override;
    virtual unsigned getNumOperands() override;
    virtual ValuePtr getOperand(unsigned i) override;
};

struct LoadInstruction:public Instruction {
    ValuePtr from;
    ValuePtr to;
    shared_ptr<GetElementPtrInstruction> gep;
    LoadInstruction(ValuePtr from, ValuePtr to, shared_ptr<BasicBlock> bb) : Instruction{InsID::Load, bb}, from{from}, to{to} {
        newUse(from.get(), this, to.get());
        //to即这条指令本身
        to->I = this;
        reg = to;
    };
    ~LoadInstruction();
    LoadInstruction(shared_ptr<GetElementPtrInstruction> gep, ValuePtr to, shared_ptr<BasicBlock> bb) : Instruction{InsID::Load, bb}, gep{gep}, from{gep->reg}, to{to} {
        newUse(gep->reg.get(), this, to.get());
        // newUse(gep->from.get(), this, to.get());
        to->I = this;
    };
    virtual void print() override;
    virtual bool replaceValue(ValuePtr target, ValuePtr newValue) override;
    virtual unsigned getNumOperands() override;
    virtual ValuePtr getOperand(unsigned i) override;
};

struct BitCastInstruction:public Instruction {
    ValuePtr from;
    TypePtr toType;
    ~BitCastInstruction();
    BitCastInstruction(ValuePtr from, ValuePtr reg, shared_ptr<BasicBlock> bb, TypePtr toType = TypePtr(new PtrType(Type::getInt8()))) : Instruction{InsID::Bitcast, bb, reg}, from{from}, toType{toType} {
        newUse(from.get(), this, reg.get());
        //reg即这条指令本身
        reg->I = this;
    }
    virtual void print() override;
    virtual bool replaceValue(ValuePtr target, ValuePtr newValue) override;
    virtual unsigned getNumOperands() override;
    virtual ValuePtr getOperand(unsigned i) override;
};

struct ExtInstruction:public Instruction {
    static int extNum;
    static ValuePtr getExtReg(TypePtr type) { return ValuePtr(new Reg(type, "ext" + to_string(extNum++))); }
    ~ExtInstruction();
    ValuePtr from;
    TypePtr to;
    bool isign;
    ExtInstruction(ValuePtr from, TypePtr to, bool isign, shared_ptr<BasicBlock> bb) : Instruction{InsID::Ext, bb, getExtReg(to)}, from{from}, to{to}, isign{isign} {
        newUse(from.get(), this, reg.get());
        //reg即这条指令本身
        reg->I = this;
    }
    virtual void print() override;
    virtual bool replaceValue(ValuePtr target, ValuePtr newValue) override;
    virtual unsigned getNumOperands() override;
    virtual ValuePtr getOperand(unsigned i) override;
};

struct SitofpInstruction:public Instruction {
    static int convNum;
    static ValuePtr getConvReg(TypePtr type) { return ValuePtr(new Reg(type, "conv" + to_string(convNum++))); }
    ValuePtr from;
    ~SitofpInstruction();
    SitofpInstruction(ValuePtr from, shared_ptr<BasicBlock> bb) : Instruction{InsID::Sitofp, bb, getConvReg(Type::getFloat())}, from{from} {
        newUse(from.get(), this, reg.get());
        //reg即这条指令本身
        reg->I = this;
    }
    virtual void print() override;
    virtual bool replaceValue(ValuePtr target, ValuePtr newValue) override;
    virtual unsigned getNumOperands() override;
    virtual ValuePtr getOperand(unsigned i) override;

    friend struct FptosiInstruction;
};

struct FptosiInstruction:public Instruction {
    ValuePtr from;
    ~FptosiInstruction();
    FptosiInstruction(ValuePtr from, shared_ptr<BasicBlock> bb) : Instruction{InsID::Fptosi, bb, SitofpInstruction::getConvReg(Type::getInt())}, from{from} {
        newUse(from.get(), this, reg.get());
        //reg即这条指令本身
        reg->I = this;
    }
    virtual void print() override;
    virtual bool replaceValue(ValuePtr target, ValuePtr newValue) override;
    virtual unsigned getNumOperands() override;
    virtual ValuePtr getOperand(unsigned i) override;
};

struct CallInstruction:public Instruction {
    static int callRegNum;
    static ValuePtr getCallReg(TypePtr type) { return ValuePtr(new Reg(type, "call" + to_string(callRegNum++))); }
    shared_ptr<Function> func;
    vector<ValuePtr> argv;
    ~CallInstruction();
    CallInstruction(shared_ptr<Function> func, vector<ValuePtr> argv, shared_ptr<BasicBlock> bb);
    virtual void print() override;
    virtual bool replaceValue(ValuePtr target, ValuePtr newValue) override;
    virtual unsigned getNumOperands() override;
    virtual ValuePtr getOperand(unsigned i) override;
};

enum BinaryInstructionOps
{
    Add,
    Mul,
    Sub,
    Div,
    Rem,
    Xor,
    Shl,
    AShr,
    FAdd,
    FSub,
    FMul,
    FDiv
};


struct BinaryInstruction : public Instruction
{
    static int BinaryRegNum;
    static ValuePtr getBinaryReg(TypePtr type) { return ValuePtr(new Reg(type, "binary" + to_string(BinaryRegNum++))); }
    ValuePtr a;
    ValuePtr b;
    bool isrv64;
    // especially ',' for '>>' and '.' for '<<'
    char op;
    ~BinaryInstruction();
    BinaryInstruction(ValuePtr a, ValuePtr b, char op, shared_ptr<BasicBlock> bb) : Instruction{InsID::Binary, bb, getBinaryReg(a->type)}, a{a}, b{b}, isrv64{false}, op{op} 
    {
        // std::cerr << "new Binary is here" << std::endl;
        newUse(a.get(), this, reg.get());
        // std::cerr << "new Binary step0" << std::endl;
        newUse(b.get(), this, reg.get());
        // std::cerr << "new Binary step1" << std::endl;
        reg->I = this;
        if (op == '!')
            reg->type = Type::getBool();

        if(a->type->isInt()&&b->type->isInt()&&(op == '+'||op == '*'||op == '!')){
            isCommutative = true;
            isAssociative = true;
        }
        if(a->type->isInt()&&b->type->isInt()&&op == '!'){
            isNilpotent = true;
        }

        // std::cerr << "new Binary step2" << std::endl;

    };

    //I为insertbefore
    BinaryInstruction(ValuePtr a, ValuePtr b, char op, Instruction* I);

    void swapOperands(){
        ValuePtr temp;
        temp = a;
        a = b; 
        b = temp;
    }
    void setOperands(unsigned index, ValuePtr newOp){
        assert(index<2&&"binaryInstruction only has two operands");
        assert(!reg->isConst&&!(dynamic_cast<Variable*>(newOp.get())&&dynamic_cast<Variable*>(newOp.get())->isGlobal)&&"cannot mutate a constant with setOperand");
        if(index == 0){
            Use *use = a->useHead;
            while(use) {
                if(use->user == this) {
                    use->rmUse();
                    break;
                }
                use = use->next;
            }
            a = newOp;
            newUse(newOp.get(),this,reg.get());
        }
        else if(index == 1){
            Use *use = b->useHead;
            while(use) {
                if(use->user == this) {
                    use->rmUse();
                    break;
                }
                use = use->next;
            }
            b = newOp;
            newUse(newOp.get(),this,reg.get());
        }
    }


    unsigned getOpcode(){
        if (a->type->isInt() || a->type->isBool())
        {
            switch (op)
            {
            case '+':
                return Add;
                break;
            case '-':
                return Sub;
                break;
            case '*':
                return Mul;
                break;
            case '/':
                return Div;
                break;
            case '%':
                return Rem;
                break;
            case '!':
                return Xor;
                break;
            case ',':
                return Shl;
                break;
            case '.':
                return AShr;
                break;

            default:
                assert(false&&"Binary type errpr");
                break;
            }
        }
        else if (a->type->isFloat())
        {
            switch (op)
            {
            case '+':
                return FAdd;
                break;
            case '-':
                return FSub;
                break;
            case '*':
                return FMul;
                break;
            case '/':
                return FDiv;
                break;
            default:
                assert(false&&"Binary Float type error");
                break;
            }
        }
        return 0;
    }

    

    virtual void print() override;
    virtual bool replaceValue(ValuePtr target, ValuePtr newValue) override;
    virtual unsigned getNumOperands() override;
    virtual ValuePtr getOperand(unsigned i) override;
};

struct FnegInstruction:public Instruction {
    static int FnegRegNum;
    static ValuePtr getFnegReg() { return ValuePtr(new Reg(Type::getFloat(), "fneg" + to_string(FnegRegNum++))); }
    ValuePtr a;
    ~FnegInstruction();
    FnegInstruction(ValuePtr a, shared_ptr<BasicBlock> bb) : Instruction{InsID::Fneg, bb, getFnegReg()}, a{a} {
        newUse(a.get(), this, reg.get());
        reg->I = this;
    }
    virtual void print() override;
    virtual bool replaceValue(ValuePtr target, ValuePtr newValue) override;
    virtual unsigned getNumOperands() override;
    virtual ValuePtr getOperand(unsigned i) override;
};

struct IcmpInstruction:public Instruction {
    static int cmpRegNum; // 这里本来应该是 block 来处理，我看 llvm ir 不会报错就偷懒没改了
    static std::map<string, IcmpKind> kindMap;
    static ValuePtr getCmpReg() { return ValuePtr(new Reg(Type::getBool(), "cmp" + to_string(cmpRegNum++))); }
    ValuePtr a;
    ValuePtr b;
    string op;
    IcmpKind kind;

    void swapOperands(){
        ValuePtr temp;
        temp = a;
        a = b; 
        b = temp;
    }

    ~IcmpInstruction();
    IcmpInstruction(shared_ptr<BasicBlock> bb, ValuePtr a, ValuePtr b = Const::getConst(Type::getInt(), 0)/*ValuePtr(new Const(Type::getInt(), 0))*/, string op = "!=") : Instruction{InsID::Icmp, bb, getCmpReg()}, a{a}, b{b}, op{op} {
        newUse(a.get(), this, reg.get());
        newUse(b.get(), this, reg.get());
        reg->I = this;
        kind = kindMap[op];
    };
    
    virtual void print() override;
    virtual bool replaceValue(ValuePtr target, ValuePtr newValue) override;
    virtual unsigned getNumOperands() override;
    virtual ValuePtr getOperand(unsigned i) override;

    friend struct FcmpInstruction;
};

struct FcmpInstruction:public Instruction {
    ValuePtr a;
    ValuePtr b;
    string op;
    ~FcmpInstruction();
    FcmpInstruction(shared_ptr<BasicBlock> bb, ValuePtr a, ValuePtr b = Const::getConst(Type::getFloat(), (float)0)/*ValuePtr(new Const(Type::getFloat(), (float)0))*/, string op = "!=") : Instruction{InsID::Fcmp, bb, IcmpInstruction::getCmpReg()}, a{a}, b{b}, op{op} {
        newUse(a.get(), this, reg.get());
        newUse(b.get(), this, reg.get());
        reg->I = this;
    };
    virtual void print() override;
    virtual bool replaceValue(ValuePtr target, ValuePtr newValue) override;
    virtual unsigned getNumOperands() override;
    virtual ValuePtr getOperand(unsigned i) override;
};

struct BrInstruction:public Instruction {
    static int ifThenNum;
    static int ifEndNum;
    static int ifElseNum;
    static int orNum;
    static int andNum;
    static int whileCondNum;
    static int whileBodyNum;
    static int whileEndNum;
    static string getifThenStr() { return "if.then" + to_string(ifThenNum++); }
    static string getifEndStr() { return "if.end" + to_string(ifEndNum++); }
    static string getifElseStr() { return "if.else" + to_string(ifElseNum++); }
    static string getOrStr() { return "lor.lhs.false" + to_string(orNum++); }
    static string getAndStr() { return "land.lhs.true" + to_string(andNum++); }
    static string getWhileCondStr() { return "while.cond" + to_string(whileCondNum++); }
    static string getwhileBodyStr() { return "while.body" + to_string(whileBodyNum++); }
    static string getwhileEndStr() { return "while.end" + to_string(whileEndNum++); }
    ValuePtr exp = nullptr;
    LabelPtr label1;
    LabelPtr label2;
    ~BrInstruction();
    BrInstruction(ValuePtr exp, LabelPtr label1, LabelPtr label2, shared_ptr<BasicBlock> bb) : Instruction{InsID::Br, bb}, exp{exp}, label1{label1}, label2{label2} {
        newUse(exp.get(), this);
    }
    BrInstruction(LabelPtr label, shared_ptr<BasicBlock> bb) : Instruction{InsID::Br, bb}, label1{label} {}
    virtual void print() override;
    virtual bool replaceValue(ValuePtr target, ValuePtr newValue) override;
    virtual unsigned getNumOperands() override;
    virtual ValuePtr getOperand(unsigned i) override;
};

struct PhiInstruction:public Instruction {
    static int phiRegNum; 
    static ValuePtr getPhiReg(TypePtr type) { return ValuePtr(new Reg(type, "phi" + to_string(phiRegNum++))); }
    ValuePtr val; // mem2reg中才会用到
    vector<std::pair<ValuePtr,shared_ptr<BasicBlock>>> from;

    string op;
    ~PhiInstruction();
    PhiInstruction(shared_ptr<BasicBlock> bb, ValuePtr val) : Instruction{InsID::Phi, bb, getPhiReg(val->type)}, val{val} {
        // newUse(val.get(), this, reg.get());
        reg->I = this;
    }; 
    //bug 无法print
    PhiInstruction(shared_ptr<BasicBlock> bb, TypePtr type) : Instruction{InsID::Phi, bb, getPhiReg(type)}, val{nullptr} {
        // newUse(val.get(), this, reg.get());
        val = Const::getConst(type,0);
        reg->I = this;
    }; 
    // addFrom单独加就行
    /*
    a   d
    |   |
    b   e
    |   |
    c---|
    |
    f
    pred 必须是c的前置基本块，这里是b和e
    */
    void addFrom(ValuePtr value, shared_ptr<BasicBlock> pred){
        from.push_back({value,pred});
        newUse(value.get(), this, reg.get());
    }
    void removeIncomingByBB(shared_ptr<BasicBlock> bb);
    virtual void print() override;
    virtual bool replaceValue(ValuePtr target, ValuePtr newValue) override;
    virtual unsigned getNumOperands() override;
    virtual ValuePtr getOperand(unsigned i) override;
};

struct LoopGEPInstruction: public Instruction {
    static int gepRegNum;
    static ValuePtr getLoopGEPReg(TypePtr type) { return ValuePtr(new Reg(type, "loopGEPReg" + to_string(gepRegNum++))); }

    GetElementPtrInstruction *gep;
    shared_ptr<Value> index;
    int initVal, stepVal;
    
    ~LoopGEPInstruction();
    LoopGEPInstruction(shared_ptr<BasicBlock> bb, GetElementPtrInstruction *_gep, shared_ptr<Value> _index, 
        int _initVal, int _stepVal): Instruction{InsID::LoopGEP, bb, getLoopGEPReg(_index->type)}, gep(_gep), index(_index), initVal(_initVal), stepVal(_stepVal) {
        _gep->loopGEP = this;
    }

    virtual void print() {}
    virtual bool replaceValue(ValuePtr target, ValuePtr newValue) { return false; }
    virtual unsigned getNumOperands() { return 0; }
    virtual ValuePtr getOperand(unsigned i) { return nullptr; }

    int getInitOffset();
    int getStepOffset();
};

struct SelectInstruction:public Instruction {
    static int selRegNum; 
    static ValuePtr getSelReg(TypePtr type) { return ValuePtr(new Reg(type, "sel" + to_string(selRegNum++))); }
    ValuePtr exp;
    ValuePtr val1,val2; 

    ~SelectInstruction();
    SelectInstruction(shared_ptr<BasicBlock> bb, ValuePtr exp, ValuePtr val1, ValuePtr val2) : Instruction{InsID::Select, bb, getSelReg(val1->type)}, exp{exp}, val1{val1}, val2{val2} {
        // newUse(val.get(), this, reg.get());
        reg->I = this;
    }; 

    virtual void print() override;
    virtual bool replaceValue(ValuePtr target, ValuePtr newValue) override;
    virtual unsigned getNumOperands() override;
    virtual ValuePtr getOperand(unsigned i) override;
};

void replaceVarByVar(ValuePtr from, ValuePtr to);
void replaceVarByVarForLCSSA(ValuePtr from, ValuePtr to, Use* use); 
void deleteUser(ValuePtr user);
void deleteUser(InstructionPtr user);
void deleteUser(Instruction* user);