#include "ADCE.h"

static unordered_set<InstructionPtr> aliveInstr; // all alive instruction
static unordered_set<ValuePtr> aliveGlobal;     // visited global variable/array
static unordered_set<ValuePtr> alivePtr;        // visited local and global variable/array and their GEP
static bool globalChanged;
static bool functionChanged;

bool isUsefulInstr(InstructionPtr instr) {
    if(instr->type == Store) {
        auto store = dynamic_cast<StoreInstruction *>(instr.get());
        if(alivePtr.count(store->des))
            return true;
    }
    else if(instr->type == Load) {
        auto load = dynamic_cast<LoadInstruction *>(instr.get());
        if(alivePtr.count(load->from))
            return true;
    }
    else if(instr->type == Call) {
        auto call = dynamic_cast<CallInstruction *>(instr.get());
        if(call->func->isLib)
            return true;
    }
    else if(instr->type == Return) {
        auto ret = dynamic_cast<ReturnInstruction *>(instr.get());
        if(!ret->retValue->isVoid)
            return true;
    }
    return false;
}

void makeAllUserAlive(ValuePtr value) {
    if(value == nullptr)
        return;
    if(value->I == nullptr) {
        alivePtr.insert(value);
        return;
    }
    else {
        auto instr = value->I->getSharedThis();
        aliveInstr.insert(instr);
        
        // record GEP and bitcast
        if(instr->type == GEP || instr->type == Bitcast) {
            alivePtr.insert(instr->reg);
        }

        if(instr->reg == nullptr)
            return;

        makeAllUserAlive(instr->reg);
    }
}

void findUsefulValue(InstructionPtr instr) {
    if(instr->type == Store) {
        auto store = dynamic_cast<StoreInstruction *>(instr.get());
        // store->des haven't been visited
        if(alivePtr.count(store->des) == 0) {
            alivePtr.insert(store->des);
            makeAllUserAlive(store->des);
        }
    }
    else if(instr->type == Load) {
        auto load = dynamic_cast<LoadInstruction *>(instr.get());
        if(alivePtr.count(load->from)) {
            alivePtr.insert(load->from);
            makeAllUserAlive(load->from);
        }
    }
}

void getUsefulValue(ValuePtr value) {
    if(auto var = dynamic_cast<Variable *>(value.get())) {
        if(var->isGlobal && aliveGlobal.count(value) == 0) {
            aliveGlobal.insert(value);
            alivePtr.insert(value);
            globalChanged = true;
        }
        else if(alivePtr.count(value) == 0) {
            alivePtr.insert(value);
        }
    }
}

static void findAliveInstr(InstructionPtr instr) {
    if(instr == nullptr || aliveInstr.count(instr))
        return;

    aliveInstr.insert(instr);
    findUsefulValue(instr);

    int numOperand = instr->getNumOperands();
    for(int i = 0; i < numOperand; i ++) {
        if(instr->getOperand(i)->I)
            findAliveInstr(instr->getOperand(i)->I->getSharedThis());
        else
            getUsefulValue(instr->getOperand(i));
    }
}

void ADCE(Module &ir) {
    aliveInstr.clear();
    aliveGlobal.clear();
    alivePtr.clear();

    do {
        globalChanged = false;
        for(auto func: ir.globalFunctions) {
            if(func->isLib)
                continue;

            do {
                functionChanged = false;
                for(int i = func->block->basicBlocks.size() - 1; i >= 0; i --) {
                    auto bb = func->block->basicBlocks[i];
                    for(int j = bb->instructions.size() - 1; j >= 0; j --) {
                        auto instr = bb->instructions[j];
                        if(isUsefulInstr(instr))
                            findAliveInstr(instr);
                    }
                }
            } while(functionChanged);
        }
    } while(globalChanged);

    for(auto func: ir.globalFunctions) {
        if(func->isLib)
            continue;

        for(auto bb: func->block->basicBlocks) {
            vector<InstructionPtr> newInstr;
            for(auto instr: bb->instructions) {
                if(aliveInstr.count(instr))
                    newInstr.push_back(instr);
            }
            bb->instructions = newInstr;
        }
    }
}
