#include "SimplifyInstructions.h"

ValuePtr simplifyInstruction(Instruction *instr)
{
    if (instr->type == Binary || instr->type == Icmp || instr->type == Fcmp)
    {
        return simplifyInstructionBinary(instr);
    }
    else if (auto phi = dynamic_cast<PhiInstruction *>(instr))
    {
        ////qaq << "phi\n";
        return simplifyInstructionPhi(phi);
    }
    else if (auto fneg = dynamic_cast<FnegInstruction *>(instr))
    {
        return simplifyInstructionFneg(fneg);
    }
    else
    {
        assert(instr->reg);
        return instr->reg;
    }
}

ValuePtr simplifyInstructionBinary(Instruction *instr)
{
    //qaq << "sim:23\n";
    if (auto c = constantFoldBinary(instr))
    {
        return c;
    }
    //qaq << "sim:32\n";
    if (instr->type == Binary)
    {
        auto bin = dynamic_cast<BinaryInstruction *>(instr);
        auto lhs = bin->a;
        auto rhs = bin->b;
        switch (bin->getOpcode())
        {
        case Add:
            if (lhs->isConst)
            {
                bin->swapOperands();
                lhs = bin->a;
                rhs = bin->b;
            }
            // X + 0 = 0
            if (rhs->isConst)
            {
                auto crhs = dynamic_cast<Const *>(rhs.get());
                if (crhs && crhs->intVal == 0)
                    return lhs;
            }
            // X + (0 - X) = 0, (0 - X) + X = 0
            if (auto slhs = dynamic_cast<BinaryInstruction *>(lhs->I); slhs && slhs->getOpcode() == Sub)
            {
                if (slhs->a->isConst)
                {
                    auto clhsofLhs = dynamic_cast<Const *>(slhs->a.get());
                    if (clhsofLhs && clhsofLhs->intVal == 0 && slhs->b == rhs)
                    {
                        auto newConst = Const::getConst(Type::getInt(), 0);
                        return newConst;
                    }
                }
            }
            else if (auto srhs = dynamic_cast<BinaryInstruction *>(rhs->I); srhs && srhs->getOpcode() == Sub)
            {
                if (srhs->a->isConst)
                {
                    auto clhsofRhs = dynamic_cast<Const *>(srhs->a.get());
                    if (clhsofRhs && clhsofRhs->intVal == 0 && srhs->b == lhs)
                    {
                        auto newConst = Const::getConst(Type::getInt(), 0);
                        return newConst;
                    }
                }
            }
            // (x - y) + (y - x) = 0
            if (auto slhs = dynamic_cast<BinaryInstruction *>(lhs->I); slhs && slhs->getOpcode() == Sub)
            {
                if (auto srhs = dynamic_cast<BinaryInstruction *>(rhs->I); srhs && srhs->getOpcode() == Sub)
                {
                    if (slhs->a == srhs->b && slhs->b == srhs->a)
                    {
                        auto newConst = Const::getConst(Type::getInt(), 0);
                        return newConst;
                    }
                }
            }
            return bin->reg;
        case Sub:
            // X - 0 = X
            if (rhs->isConst)
            {
                if (auto crhs = dynamic_cast<Const *>(rhs.get()); crhs && crhs->intVal == 0)
                {
                    return lhs;
                }
            }
            // 0 - (0 - X) = X
            if (lhs->isConst)
            {
                if (auto clhs = dynamic_cast<Const *>(lhs.get()); clhs && clhs->intVal == 0)
                {
                    if (auto srhs = dynamic_cast<BinaryInstruction *>(rhs.get()); srhs && srhs->getOpcode() == Sub)
                    {
                        if (srhs->a->isConst)
                        {
                            if (auto clhsofRhs = dynamic_cast<Const *>(srhs->a.get()); clhsofRhs && clhsofRhs->intVal == 0)
                            {
                                return srhs->b;
                            }
                        }
                    }
                }
            }
            // X - X = 0
            if (lhs == rhs)
            {
                auto newConst = Const::getConst(Type::getInt(), 0);
                return newConst;
            }
            return bin->reg;
        case Mul:
            if (lhs->isConst)
            {
                bin->swapOperands();
                lhs = bin->a;
                rhs = bin->b;
            }
            // X * 0 = 0 X * 1 = 1
            if (rhs->isConst)
            {
                if (auto crhs = dynamic_cast<Const *>(rhs.get()); crhs)
                {
                    if (crhs->intVal == 0)
                    {
                        auto newConst = Const::getConst(Type::getInt(), 0);
                        return newConst;
                    }
                    else if (crhs->intVal == 1)
                    {
                        return lhs;
                    }
                }
            }
            return bin->reg;
        case Div:
            // X / 1 = X
            if (rhs->isConst)
            {
                if (auto crhs = dynamic_cast<Const *>(rhs.get()); crhs)
                {
                    if (crhs->intVal == 1)
                    {
                        return lhs;
                    }
                }
            }
            return bin->reg;
//        case FAdd:
//            return bin->reg.get();
//        case FSub:
//            return bin->reg.get();
//        case FMul:
//            return bin->reg.get();
//        case FDiv:
//            return bin->reg.get();
        default:
            return bin->reg;
        }
    }
    else if (instr->type == Icmp)
    {
        auto bin = dynamic_cast<IcmpInstruction *>(instr);
        //qaq << "sim:177\n";
        auto lhs = bin->a;
        auto rhs = bin->b;
        if (bin->op == "==")
        {
            if (lhs == rhs)
            {
                // cmp应当返回什么
                auto newConst = Const::getConst(Type::getBool(), true);
                return newConst;
            }
/*            if (lhs->isConst)
            {
                bin->swapOperands();
                lhs = bin->a;
                rhs = bin->b;
            }
            */
           else
           {
                return bin->reg;
           }
        }
        else
        {
            return bin->reg;
        }
    }
    else if (instr->type == Fcmp)
    {
        return instr->reg;
    }
    else
    {
        assert(0);
    }
}

ValuePtr simplifyInstructionPhi(PhiInstruction *instr)
{
    bool same = true;
    auto first = instr->from[0].first;
    for (auto From : instr->from)
    {
        if (From.first != first)
        {
            same = false;
            break;
        }
    }
    if (same)
        return first;
    return instr->reg;
}

ValuePtr simplifyInstructionFneg(FnegInstruction *instr)
{
    if (auto c = constantFoldFneg(instr))
        return c;

    auto val = instr->reg;
    auto bin = new BinaryInstruction(Const::getConst(Type::getFloat(), (float)0.0),
                                     val, '-', instr->basicblock);
    return bin->reg;
}

ValuePtr constantFoldFneg(FnegInstruction *instr)
{
    if (instr->a->isConst)
    {
        assert(instr->a->type->isFloat());
        auto c = dynamic_cast<Const *>(instr->a.get());
        auto myres = -c->floatVal;
        auto res = Const::getConst(Type::getFloat(), myres);
        return res;
    }
    else
    {
        return nullptr;
    }
}

ValuePtr constantFoldBinary(Instruction *instr)
{
    /*if (!(instr->a->isConst && instr->b->isConst))
        return nullptr;

    auto clhs = dynamic_cast<Const*>(instr->a.get());
    auto crhs = dynamic_cast<Const*>(instr->b.get());
    if(instr->a->type->isInt() && instr->b->type->isInt()) // 两边都是整数
    {
        long long x = clhs->intVal, y = crhs->intVal;
        switch (instr->type)
        {
        case BinaryInstructionOps::Add:
            Return new Const::getConst(Type::getInt(), x + y);
        case
        }
    }
    else if(instr->a->type->isFloat() || instr->b->type->isFloat()) // 有一边是浮点数
    {

    }
    else
    {
        return nullptr;
    }*/
    if (instr->type == Binary)
    {
        auto Bi = (BinaryInstruction *)(instr);
        if (Bi->a->isConst && Bi->b->isConst)
        {
            //qaq << "sim:295\n";
            auto op = Bi->op;
            auto lhs = dynamic_cast<Const *>(Bi->a.get());
            auto rhs = dynamic_cast<Const *>(Bi->b.get());
            if (Bi->a->type->isFloat() || Bi->b->type->isFloat())
            {
                float myres = 0.0;
                float lhs_value, rhs_value;
                if (Bi->a->type->isFloat())
                    lhs_value = lhs->floatVal;
                else
                    lhs_value = lhs->intVal * 1.0;
                if (Bi->b->type->isFloat())
                    rhs_value = rhs->floatVal;
                else
                    rhs_value = rhs->intVal * 1.0;

                if (op == '+')
                    myres = lhs_value + rhs_value;
                else if (op == '-')
                    myres = lhs_value - rhs_value;
                else if (op == '*')
                    myres = lhs_value * rhs_value;
                else if (op == '/')
                    myres = lhs_value / rhs_value;
                else
                    return nullptr;

                auto res = Const::getConst(Type::getFloat(), (float)myres);
                return res;
            }
            else if (Bi->a->type->isBool() || Bi->b->type->isBool())
            {
                bool myres = false;
                bool lhs_value, rhs_value;

                lhs_value = lhs->boolVal;
                rhs_value = rhs->boolVal;
                if (op == '+')
                    myres = lhs_value + rhs_value;
                else if (op == '-')
                    myres = lhs_value - rhs_value;
                else if (op == '*')
                    myres = lhs_value * rhs_value;
                else if (op == '/')
                    myres = lhs_value / rhs_value;
                else if (op == '%')
                    myres = lhs_value % rhs_value;
                else if (op == '!')
                    myres = !lhs_value;
                else if (op == ',')
                    myres = lhs_value << rhs_value;
                else if (op == '.')
                    myres = lhs_value >> rhs_value;
                else
                    return nullptr;
                auto res = Const::getConst(Type::getBool(), (bool)myres);

                return res;
            }
            else
            {
                int myres = 0.0;
                int lhs_value, rhs_value;

                // lhs_value = lhs->intVal;
                // rhs_value = rhs->intVal;
                lhs_value = lhs->type->isBool() ? lhs->boolVal : lhs->intVal;
                rhs_value = rhs->type->isBool() ? rhs->boolVal : rhs->intVal;
                if (op == '+')
                    myres = lhs_value + rhs_value;
                else if (op == '-')
                    myres = lhs_value - rhs_value;
                else if (op == '*')
                    myres = lhs_value * rhs_value;
                else if (op == '/')
                    myres = lhs_value / rhs_value;
                else if (op == '%')
                    myres = lhs_value % rhs_value;
                else if (op == '!')
                    // myres = lhs_value ^ rhs_value;
                    myres = !lhs_value;
                else if (op == ',')
                    myres = lhs_value << rhs_value;
                else if (op == '.')
                    myres = lhs_value >> rhs_value;
                else
                    return nullptr;
                //qaq << myres << endl;
                auto res = Const::getConst(Type::getInt(), (int)myres);

                return res;
            }
        }
        else
        {
            return nullptr;
        }
    }
    else if (instr->type == Icmp)
    {
        //qaq << "Icmp\n";
        auto Ii = dynamic_cast<IcmpInstruction *>(instr);
        ////qaq << "here\n";
        if (Ii->a->isConst && Ii->b->isConst)
        {
            //qaq << "here\n";
            bool res = false;
            // 不可能是float，如果是，那就有bug
            assert(!Ii->a->type->isFloat() && !Ii->b->type->isFloat() && "Icmp with float value");
            int aVal = Ii->a->type->isInt() ? dynamic_cast<Const *>(Ii->a.get())->intVal : dynamic_cast<Const *>(Ii->a.get())->boolVal;
            int bVal = Ii->b->type->isInt() ? dynamic_cast<Const *>(Ii->b.get())->intVal : dynamic_cast<Const *>(Ii->b.get())->boolVal;
            string op = Ii->op;
            if (op == "!=")
            {
                res = aVal != bVal;
            }
            else if (op == ">")
            {
                res = aVal > bVal;
            }
            else if (op == ">=")
            {
                res = aVal >= bVal;
            }
            else if (op == "<")
            {
                res = aVal < bVal;
            }
            else if (op == "<=")
            {
                res = aVal <= bVal;
            }
            else if (op == "==")
            {
                res = aVal == bVal;
            }
            else
            {
                assert(false && "wrong op");
            }
            auto newConst = Const::getConst(Type::getBool(), res);

            return newConst;
        }
        else
        {
            //qaq << "null\n";
            return nullptr;
        }
    }
    else if (instr->type == Fcmp)
    {
        auto Fi = dynamic_cast<FcmpInstruction *>(instr);
        if (Fi->a->isConst && Fi->b->isConst)
        {
            bool res = false;
            // 不可能是int，如果是，那就有bug
            assert(!Fi->a->type->isInt() && !Fi->b->type->isInt() && "Icmp with float value");
            double aVal = Fi->a->type->isFloat() ? dynamic_cast<Const *>(Fi->a.get())->floatVal : dynamic_cast<Const *>(Fi->a.get())->boolVal;
            double bVal = Fi->b->type->isFloat() ? dynamic_cast<Const *>(Fi->b.get())->floatVal : dynamic_cast<Const *>(Fi->b.get())->boolVal;
            string op = Fi->op;
            if (op == "!=")
            {
                res = aVal != bVal;
            }
            else if (op == ">")
            {
                res = aVal > bVal;
            }
            else if (op == ">=")
            {
                res = aVal >= bVal;
            }
            else if (op == "<")
            {
                res = aVal < bVal;
            }
            else if (op == "<=")
            {
                res = aVal <= bVal;
            }
            else if (op == "==")
            {
                res = aVal == bVal;
            }
            else
            {
                assert(false && "wrong op");
            }
            auto newConst = Const::getConst(Type::getBool(), res);

            return newConst;
        }
        else
        {
            return nullptr;
        }
    }
    else
    {
        assert(0);
    }
}