#include "local2Global.h"

static vector<ValuePtr> getLocalArr(BlockPtr block) {
    auto entry = block->getEntryBlock();
    vector<ValuePtr> localArr;
    for(auto instr: entry->instructions) {
        if(instr->type == Alloca) {
            auto alloc = dynamic_cast<AllocaInstruction *>(instr.get());
            if(alloc->des->type->isArr())
                localArr.emplace_back(alloc->des);
        }
    }
    return localArr;
}

static bool isLocalArrStoreBeforeLoad(ValuePtr arr, FunctionPtr func) {
    bool isLoad = false;
    // record arr or GEPs from arr
    unordered_set<ValuePtr> relatedGEP;
    relatedGEP.insert(arr);
    for(auto bb: func->block->basicBlocks) {
        for(auto instr: bb->instructions) {
            if(instr->type == Load) {
                auto load = dynamic_cast<LoadInstruction *>(instr.get());
                if(relatedGEP.count(load->from)) {
                    isLoad = true;
                }
            }
            // if store after some load, return false
            else if(instr->type == Store) {
                auto store = dynamic_cast<StoreInstruction *>(instr.get());
                if(relatedGEP.count(store->des) && isLoad) {
                    return false;
                }
            }
            else if(instr->type == GEP) {
                auto gep = dynamic_cast<GetElementPtrInstruction *>(instr.get());
                if(relatedGEP.count(gep->from)) {
                    relatedGEP.insert(gep->reg);
                }
            }
            else if(instr->type == Bitcast) {
                auto cast = dynamic_cast<BitCastInstruction *>(instr.get());
                if(relatedGEP.count(cast->from)) {
                    relatedGEP.insert(cast->reg);
                }
            }
            // if GEP is passed as function parameter, return false
            else if(instr->type == Call) {
                auto call = dynamic_cast<CallInstruction *>(instr.get());
                for(auto arg: call->argv) {
                    if(relatedGEP.count(arg))
                        return false;
                }
            }
        }
    }
    return true;
}

static vector<int> getShape(TypePtr type) {
    if(type->isArr()) {
        auto arrType = dynamic_cast<ArrType *>(type.get());
        vector<int> shape{arrType->length};
        vector<int> innerShape = getShape(arrType->inner);
        for(auto index: innerShape)
            shape.emplace_back(index);

        return shape;
    }
    else if(type->isInt()) {
        return vector<int>{};
    }
    else if(type->isFloat()) {
        return vector<int>{};
    }
    else {
        fprintf(stderr, "type %s is not supported in getShape!", type->getStr().c_str());
        assert(false && "type is not supported in getShape!");
    }
}

static void dfsUseVisit(InstructionPtr instr, function<bool(InstructionPtr)> visit, function<void(InstructionPtr)> action) {
    if(visit(instr))
        return;

    action(instr);
    auto reg = instr->type == Alloca ? dynamic_cast<AllocaInstruction *>(instr.get())->des: instr->reg;
    if(reg == nullptr || (instr->type != GEP && instr->type != Alloca && instr->type != Bitcast))
        return;
    auto use = reg->useHead;
    while(use) {
        if(use->user)
            dfsUseVisit(use->user->getSharedThis(), visit, action);
        use = use->next;
    }
}

bool local2Global(Module &ir) {
    printf("enter local2Global\n");
    fflush(stdout);
    bool fff =false;
    for(auto func: ir.globalFunctions) {
        unordered_set<InstructionPtr> toDelete;
        if(func->isLib)
            continue;
        // printf("function %s\n", func->name.c_str());
        // fflush(stdout);
        vector<ValuePtr> localArr = getLocalArr(func->block);
        // printf("finish get local arrays: ");
        // for(auto arr: localArr)
        //     printf("%s, ", arr->getStr().c_str());
        // printf("\n");
        fflush(stdout);

        for(auto arr: localArr) {
            assert(arr->type->isArr());
            // only take load after store into considerasion
            if(!isLocalArrStoreBeforeLoad(arr, func))
                continue;
            vector<int> shape = getShape(arr->type);
            if(shape.size() != 1)
                continue;

            printf("shape: [%d]\n", shape[0]);

            // only handle one-dimension int array
            unordered_map<ValuePtr, int> valIndexMap;
            vector<bool> visitedArr(shape[0], false);   // is a position visited
            vector<int> constValues(shape[0], 0);       // initial value of array
            vector<vector<StoreInstruction *>> constStoreInstr(shape[0], vector<StoreInstruction *>{});
            unordered_map<ValuePtr, bool> GEPMutable;   // is GEP mutable
            unordered_set<InstructionPtr> visitedInstr; // is an instruction visited
            unordered_set<StoreInstruction *> storeInstr;   // tracked store instr
            unordered_set<LoadInstruction *> loadInstr;   // tracked load instr
            unordered_set<GetElementPtrInstruction *> gepInstr;   // tracked GEP instr
            bool flag = true;
            GEPMutable[arr] = false;

            // dfs traverse for arr uses
            dfsUseVisit(arr->I->getSharedThis(), [&](InstructionPtr instr) -> bool {
                if(visitedInstr.count(instr))
                    return true;
                visitedInstr.insert(instr);
                return false;
            }, [&](InstructionPtr instr) -> void {
                if(!flag || instr == nullptr)
                    return;
                // instr->print();
                // fflush(stdout);
                if(instr->type == Store) {
                    auto store = dynamic_cast<StoreInstruction *>(instr.get());
                    auto gep = store->des;
                    // store->print();
                    // fflush(stdout);
                    assert(gep->I->type == GEP);
                    // if arr[index] has been stored or stored value is not a constant
                    if(!store->value->isConst || GEPMutable[store->des]) {
                        // printf("flag = false, visitedArr[%d] = %d, value->isConst: %d, GEPMutable: %d", index, visitedArr[index], store->value->isConst, GEPMutable[store->des]);
                        flag = false;
                    }
                    else {
                        assert(valIndexMap.count(gep));
                        int index = valIndexMap[gep];
                        auto constant = dynamic_cast<Const *>(store->value.get());
                        assert(constant != nullptr);
                        storeInstr.insert(store);
                        // constValues[index] = constant->intVal;
                        constStoreInstr[index].push_back(store);
                    }
                }
                else if(instr->type == Load) {
                    auto load = dynamic_cast<LoadInstruction *>(instr.get());
                    auto gep = load->from;
                    assert(gep->I->type == GEP);
                    // if load variable GEP, flag = false
                    if(!valIndexMap.count(gep)) {
                        flag = false;
                        return;
                    }
                    int index = valIndexMap[gep];
                    // if arr[index] has been stored or stored value is not a constant
                    if(GEPMutable[load->from]) {
                        // printf("flag = false, visitedArr[%d] = %d, value->isConst: %d, GEPMutable: %d", index, visitedArr[index], store->value->isConst, GEPMutable[store->des]);
                        flag = false;
                    }
                    else {
                        loadInstr.insert(load);
                    }
                }
                else if(instr->type == GEP) {
                    auto gep = dynamic_cast<GetElementPtrInstruction *>(instr.get());
                    // vector<int> prevIndex = valIndexMap[gep->from];
                    // gep->print();
                    ValuePtr index = nullptr;
                    // pointer GEP
                    if(gep->index.size() == 1) {
                        index = gep->index[0];
                    }
                    // array GEP
                    else {
                        assert(gep->from == arr);
                        assert(gep->index.size() == 2);
                        index = gep->index[1];
                    }
                    // constant index
                    if(auto constant = dynamic_cast<Const *>(index.get()); constant) {
                        valIndexMap[gep->reg] = constant->intVal;
                        GEPMutable[gep->reg] = GEPMutable[gep->from];
                    }
                    // index is not a constant
                    else {
                        GEPMutable[gep->reg] = true;
                    }
                    gepInstr.insert(gep);
                }
                // should not be function parameter
                else if(instr->type == Call) {
                    flag = false;
                }
            });

            // if 'arr' satisfy all conditions
            if(flag) {
                // printf("found local array: ");
                // arr->I->print();
                // printf("stores to be eliminated:\n");
                // for(int i = 0; i < shape[0]; i ++) {
                //     printf("index %d stores:\n", i);
                //     for(auto store: constStoreInstr[i]) {
                //         store->print();
                //     }
                // }
                fflush(stdout);

                // sort all stores
                bool isStoreOrdered = true;
                for(int i = 0; i < shape[0]; i ++) {
                    // no store to 'arr[i]'
                    if(constStoreInstr[i].size() == 0)
                        continue;

                    StoreInstruction *finalStore = constStoreInstr[i][0];
                    for(int j = 1; j < constStoreInstr[i].size(); j ++) {
                        // can not be sorted
                        if(!isStoreOrdered)
                            break;

                        StoreInstruction *curStore = constStoreInstr[i][j];
                        auto curBB = curStore->basicblock;
                        auto finalBB = finalStore->basicblock;
                        // curBB dominates finalBB
                        bool curDomFinal = false;
                        // stores in different basic blocks
                        if(curBB != finalBB) {
                            // finalBB -> curBB
                            if(curBB->allDoms.count(finalBB)) {
                                curDomFinal = false;
                            }
                            // curBB -> finalBB
                            else if(finalBB->allDoms.count(curBB)) {
                                curDomFinal = true;
                            }
                            else {
                                isStoreOrdered = false;
                                break;
                            }
                        }
                        // stores in the same basic blocks
                        else {
                            bool foundOne = false;
                            for(auto instr: curBB->instructions) {
                                // curStore -> finalStore
                                if(instr == curStore->getSharedThis()) {
                                    curDomFinal = false;
                                    foundOne = true;
                                    break;
                                }
                                // finalStore -> curStore
                                else if(instr == finalStore->getSharedThis()){
                                    curDomFinal = true;
                                    foundOne = true;
                                    break;
                                }
                            }
                            assert(foundOne);
                        }

                        if(!curDomFinal) {
                            finalStore = curStore;
                        }
                    }

                    // store constant value to compiler array
                    auto constVal = dynamic_cast<Const *>(finalStore->value.get());
                    assert(constVal && constVal->type->isInt());
                    constValues[i] = constVal->intVal;
                }

                // printf("const value in array[%d]: ", shape[0]);
                // for(int i = 0; i < shape[0]; i ++) {
                //     printf("%d, ", constValues[i]);
                // }
                // printf("\n");

                // replace loads with constant value
                for(auto load: loadInstr) {
                    assert(valIndexMap.count(load->from));
                    int index = valIndexMap[load->from];
                    auto constVal = Const::getConst(Type::getInt(), constValues[index], "local_arrary_propagation" + to_string(index));
                    replaceVarByVar(load->reg, constVal);
                    deleteUser(load->reg);
                    toDelete.insert(load->getSharedThis());
                }
                // delete all store
                for(auto store: storeInstr) {
                    toDelete.insert(store->getSharedThis());
                }
                // delete all GEP
                for(auto gep: gepInstr) {
                    toDelete.insert(gep->getSharedThis());
                }
                // delete arr allocation
                toDelete.insert(arr->I->getSharedThis());
            }
        }
        if(toDelete.size() > 0) {
            fff = true;
            // printf("deleting local array!!!\n");
            for(auto bb: func->block->basicBlocks) {
                vector<InstructionPtr> newInstr;
                for(auto instr: bb->instructions) {
                    if(toDelete.count(instr) == 0)
                        newInstr.push_back(instr);
                }
                bb->instructions = newInstr;
            }
        }
    }
    return fff;
}