#include "loopAnalysis.h"

static int loopCnt = 0;

void discoverAndMapSubloop(LoopPtr loop, BlockPtr block, stack<BasicBlockPtr>& BackedgeTo) {
    while(!BackedgeTo.empty()) {
        auto pred = BackedgeTo.top();
        BackedgeTo.pop();
        auto subLoop = block->getLoopOfBB(pred);
        if(!subLoop) {
            block->addBBToLoop(pred, loop);
            if(pred == loop->getHeader())
                continue;
            for(auto predOfPred : pred->getPredecessor())
                BackedgeTo.push(predOfPred);
        }
        else {
            while(auto parent = subLoop->getParentLoop())
                subLoop = parent;
            if(subLoop == loop)
                continue;
            subLoop->setParentLoop(loop);
            pred = subLoop->getHeader();
            for(auto predOfPred : pred->getPredecessor())
                if(!block->getLoopOfBB(predOfPred) 
                    || block->getLoopOfBB(predOfPred) != subLoop)
                    BackedgeTo.push(predOfPred);
        }
    }
}

int getLoopDepth(BasicBlockPtr bb) {
    auto func = bb->getParent();
    if(auto loop = func->block->getLoopOfBB(bb))
        return loop->getLoopDepth();
    return 0;
}

void computeExitingBlocks(BlockPtr block) {
    // cerr << "compute exiting blocks" << endl;
    for (auto loop : block->getAllLoops())  {
        for (auto bb : loop->getLoopBasicBlocks())  {
            // auto lastInstr = *(--bb->getInstructions().end());
            auto lastInstr = bb->instructions.back();
            if(auto br = dynamic_cast<BrInstruction *>(lastInstr.get()); br && br->exp) {
                // auto bb1 = br->getTrueBlock();
                // auto bb2 = br->getFalseBlock();
                // make sure that we matain block->LabelBBMap after createCFG!
                auto bb1 = block->LabelBBMap[br->label1];
                auto bb2 = block->LabelBBMap[br->label2];

                if(!loop->contains(bb1) || !loop->contains(bb2))
                {
                    loop->addExitingBlocks(bb);
                    // cerr << "Adding Exiting basicblock" << bb->label->getStr() << endl;
                }
                    // loop->getExitingBlocks().insert(bb);
                    
            }
        }

    }

}

void computeExitBlocks(BlockPtr block) {

    for (auto loop : block->getAllLoops()) 
        for (auto bb : loop->getLoopBasicBlocks()) 
            for (auto succBB : bb->getSuccessor()) 
            {
                
                if (!loop->contains(succBB))
                {
                    loop->addExitBlocks(succBB);
                    // cerr << "Adding Exit basicblock" << succBB->label->getStr() << endl;
                }
            }
                    // loop->getExitBlocks().insert(succBB);
                    
}

void computePreHeader(BlockPtr block) {
    for (auto loop : block->getAllLoops()) {
        BasicBlockPtr preHeader = nullptr;
        for (auto pred : loop->getHeader()->getPredecessor()) {
            if (getLoopDepth(pred) != loop->getLoopDepth())  {
                preHeader = pred;
                break;
            }
        }
        assert(preHeader != nullptr && preHeader->getSuccessor().size() == 1);
        loop->setPreheader(preHeader);
    }
}

void computeLatch(BlockPtr block) {
    for (auto loop : block->getAllLoops()) {
        BasicBlockPtr latch = nullptr;
        for (auto pred : loop->getHeader()->getPredecessor()) {
            if (loop->contains(pred)) {
                latch = pred;
                break;
            }
        }
        loop->setLatchBlock(latch);
    }
}

void computeLoopIndVar(BlockPtr block) {
    for (auto loop : block->getAllLoops()) {
        auto header = loop->getHeader();
        // auto lastInstr = *(--latch->end());
        auto lastInstr = header->instructions.back();
        // PrintMsg(lastInstr->getSignature());
        if(auto br = dynamic_cast<BrInstruction *>(lastInstr.get())) {
            // cerr << "Br enter" << endl;
            if(br->exp == nullptr)
            {
                // cerr << "Return 1" << endl;
                continue;
            }
                
            auto icmp = dynamic_cast<IcmpInstruction *>(br->exp->I);
            if(!icmp)
            {
                // cerr << "Return 2" << endl;
                continue;
            }
            loop->setIndexCondInstr(icmp->getSharedThis());
            loop->setICmpKind(icmp->kind);

            // cerr << "Br continue" <<endl;

            function<PhiInstruction*(Instruction*)> getPhi = [&](Instruction* instr) -> PhiInstruction* {
                if(auto phi = dynamic_cast<PhiInstruction*>(instr))
                    return phi;
                else if(auto binary = dynamic_cast<BinaryInstruction*>(instr)) {
                    auto lhs = binary->getOperand(0);
                    auto rhs = binary->getOperand(1);
                    if(auto defInstr = lhs->I; defInstr 
                        && !loop->isSimpleLoopInvariant(lhs))
                        if(auto phi = getPhi(defInstr))
                            return phi;
                    else if(auto defInstr = rhs->I; defInstr 
                        && !loop->isSimpleLoopInvariant(rhs))
                        if(auto phi = getPhi(defInstr))
                            return phi;
                }
                return nullptr;
            };
            auto icmpLhs = icmp->getOperand(0);
            auto icmpRhs = icmp->getOperand(1);
            if(loop->isSimpleLoopInvariant(icmpLhs)) {
                loop->setIndexEnd(icmpLhs);
                if(auto instr = icmpRhs->I; instr 
                    && !loop->isSimpleLoopInvariant(icmpRhs)) {
                    auto phi = getPhi(instr);
                    // cerr << "setset" <<endl;
                    if(phi)
                        loop->setIndexPhi(phi);
                }
            }
            else if(loop->isSimpleLoopInvariant(icmpRhs)) {
                // dumpBasickBlock(loop->getHeader());
                loop->setIndexEnd(icmpRhs);
                if(auto instr = icmpLhs->I; instr 
                    && !loop->isSimpleLoopInvariant(icmpLhs)) {
                    // PrintMsg(instr->getSignature());
                    auto phi = getPhi(instr);
                    // cerr << "setset" <<endl;
                    if(phi)
                        loop->setIndexPhi(phi);
                }
            }
        }
    }
}

void computeLoopBB(BlockPtr block) {
    for(auto bb: block->basicBlocks) {
        bb->loop = block->getLoopOfBB(bb);
        if(bb->loop)
            bb->loopDepth = bb->loop->getLoopDepth();
    }
}


void loopAnalysis(BlockPtr block) {
    block->clearBBToLoop();
    block->clearLoops();

    vector<BasicBlockPtr> postOrderList;
    set<BasicBlockPtr> visited;
    // DFS for dominator Tree to find loops
    BasicBlock::domTreeDFSPost(block->getEntryBlock(), [&](BasicBlockPtr bb) -> bool {
        if(visited.count(bb))
            return true;
        visited.insert(bb);
        return false;
    }, [&](BasicBlockPtr bb) -> void {
        postOrderList.push_back(bb);
    });

    stack<BasicBlockPtr> BackedgeTo;
    for(auto header : postOrderList) {
        for(auto pred : header->getPredecessor())
            if(pred->getAllDoms().count(header)) {
                BackedgeTo.push(pred);
            }
        
        if(!BackedgeTo.empty()) {
            auto loop = LoopPtr(new Loop(header, loopCnt++));
            discoverAndMapSubloop(loop, block, BackedgeTo);
        }
    }

    // DFS for basic blocks to get subloops
    visited.clear();
    BasicBlock::basicBlockDFSPost(block->getEntryBlock(), [&](BasicBlockPtr bb) -> bool {
        if(visited.count(bb))
            return true;
        visited.insert(bb);
        return false;
    }, [&](BasicBlockPtr bb) -> void {
        auto subLoop = block->getLoopOfBB(bb);
        if (subLoop && bb == subLoop->getHeader())  {
            if (subLoop->getParentLoop())
                // subLoop->getParentLoop()->getSubLoops().push_back(subLoop);
                subLoop->getParentLoop()->addSubLoop(subLoop);
            else {
                // func->getTopLoops().push_back(subLoop);
                block->addTopLoop(subLoop);
                subLoop->setLoopDepth(1); // topLoop is 1
            }

            // reverse but no header
            reverse(subLoop->getLoopBasicBlocksRef().begin() + 1, subLoop->getLoopBasicBlocksRef().end());
            reverse(subLoop->getSubLoopsRef().begin(), subLoop->getSubLoopsRef().end());
            subLoop = subLoop->getParentLoop();
        }

        for(; subLoop; subLoop = subLoop->getParentLoop()) {
            // subLoop->getLoopBasicBlocks().push_back(bb);
            // subLoop->BBSet.insert(bb);
            subLoop->addBasicBlock(bb);
        }
    });

    stack<LoopPtr, vector<LoopPtr>> loopStk(block->getTopLoops());
    while (!loopStk.empty()) {
        auto loop = loopStk.top();
        loopStk.pop();
        // func->getAllLoops().push_back(loop);
        block->addLoop(loop);
        for(auto subLoop : loop->getSubLoops()) {
            subLoop->setLoopDepth(loop->getLoopDepth() + 1);
            loopStk.push(subLoop);
        }
    }

    computeExitingBlocks(block);
    computeExitBlocks(block);
    computeLatch(block);
    computePreHeader(block);
    computeLoopIndVar(block);
    computeLoopBB(block);

    for(auto loop: block->getTopLoops())
    {
        ScalarEvolution(loop);
    }

}


    



